import os
import sys
from optparse import OptionGroup, OptionParser

import torch

from sgn.apps import Pipeline
from sgnligo.base.utils import parse_list_to_dict
from sgnligo.cbc.sort_bank import group_and_read_banks, SortedBank
from sgnligo.sinks import ImpulseSink, StrikeSink, KafkaSink
from sgnligo.sources import datasource
from sgnligo.transforms import (
    condition,
    lloid,
    Itacacac,
    Latency,
)
from sgnts.sinks import FakeSeriesSink


def parse_command_line():
    parser = OptionParser()

    group = OptionGroup(parser, "Data source", "Options for data source.")
    group.add_option(
        "--data-source",
        action="store",
        default="frames",
        help="The type of the input source. Supported sources: 'white', 'sin', "
             "'impulse', 'frames', 'devshm'",
    )
    group.add_option(
        "--channel-name",
        metavar="ifo=channel-name",
        action="append",
        help="Name of the data channel to analyze. Can be given multiple times as "
             "--channel-name=IFO=CHANNEL-NAME",
    )
    group.add_option(
        "--sample-rate",
        metavar="Hz",
        type=int,
        help="Requested sampling rate of the data.",
    )
    group.add_option(
        "--source-buffer-duration",
        type=float,
        metavar="seconds",
        action="store",
        help="Source elements will produce buffers in strides of this duration.",
    )
    group.add_option(
        "--frame-cache",
        metavar="file",
        help="Set the path to the frame cache file to analyze.",
    )
    group.add_option(
        "--gps-start-time",
        metavar="seconds",
        help="Set the start time of the segment to analyze in GPS seconds. "
             "For frame cache data source",
    )
    group.add_option(
        "--gps-end-time",
        metavar="seconds",
        help="Set the end time of the segment to analyze in GPS seconds. "
             "For frame cache data source",
    )
    group.add_option(
        "--shared-memory-dir",
        metavar="ifo=directory",
        action="append",
        help="Set the name of the shared memory directory. "
             "Can be given multiple times as --shared-memory-dir=IFO=DIR-NAME",
    )
    group.add_option(
        "--state-channel-name",
        metavar="ifo=channel-name",
        action="append",
        help="Set the state vector channel name. "
             "Can be given multiple times as --state-channel-name=IFO=CHANNEL-NAME",
    )
    group.add_option(
        "--state-vector-on-bits",
        metavar="ifo=number",
        action="append",
        help="Set the state vector on bits. "
             "Can be given multiple times as --state-vector-on-bits=IFO=NUMBER",
    )
    group.add_option(
        "--wait-time",
        metavar="seconds",
        type=int,
        default=60,
        help="Time to wait for new files in seconds before throwing an error. "
             "In online mode, new files should always arrive every second, unless "
             "there are problems. Default wait time is 60 seconds.",
    )
    group.add_option(
        "--num-buffers",
        type="int",
        action="store",
        default=10,
        help="Number of buffers the source element should produce when source is "
             "fake source",
    )
    group.add_option(
        "--impulse-position",
        action="store",
        default=-1,
        help="The sample point to put the impulse at. If -1, place randomly.",
    )
    group.add_option(
        "--impulse-ifo",
        action="store",
        help="Only do impulse test on data from this ifo.",
    )
    parser.add_option_group(group)

    group = OptionGroup(
        parser, "PSD Options", "Adjust noise spectrum estimation parameters"
    )
    group.add_option(
        "--whitening-method",
        metavar="algorithm",
        default="gstlal",
        help="Algorithm to use for whitening the data. Supported options are 'gwpy' "
             "or 'gstlal'. Default is gstlal.",
    )
    group.add_option(
        "--psd-fft-length",
        action="store",
        type=int,
        help="The fft length for psd estimation.",
    )
    group.add_option(
        "--reference-psd",
        metavar="file",
        help="load the spectrum from this LIGO light-weight XML file (optional).",
    )
    group.add_option(
        "--track-psd",
        action="store_true",
        help="Enable dynamic PSD tracking.  Always enabled if --reference-psd is not "
             "given.",
    )
    parser.add_option_group(group)

    group = OptionGroup(parser, "Data Qualtiy", "Adjust data quality handling")
    group.add_option(
        "--ht-gate-threshold",
        action="store",
        type=float,
        default=float("+inf"),
        help="The gating threshold. Data above this value will be gated out.",
    )
    parser.add_option_group(group)

    group = OptionGroup(
        parser, "Trigger Generator", "Adjust trigger generator behaviour"
    )
    group.add_option(
        "--svd-bank",
        metavar="filename",
        action="append",
        default=[],
        help="Set the name of the LIGO light-weight XML file from which to load the "
             "svd bank for a given instrument.  To analyze multiple instruments, "
             "--svd-bank can be called multiple times for svd banks corresponding "
             "to different instruments.  If --data-source is lvshm or framexmit, "
             "then only svd banks corresponding to a single bin must be given. "
             "If given multiple times, the banks will be processed one-by-one, in "
             "order.  At least one svd bank for at least 2 detectors is required, "
             "but see also --svd-bank-cache.",
    )
    group.add_option(
        "--nbank-pretend",
        type="int",
        action="store",
        default=0,
        help="Pretend we have this many subbanks by copying the first subbank "
             "this many times",
    )
    group.add_option(
        "--nslice",
        type="int",
        action="store",
        default=-1,
        help="Only filter this many timeslices. Default: -1, filter all timeslices.",
    )
    group.add_option(
        "--trigger-finding-length",
        type="int",
        metavar="samples",
        action="store",
        default=2048,
        help="Produce triggers in blocks of this many samples.",
    )
    group.add_option(
        "--impulse-bank",
        metavar="filename",
        action="store",
        default=None,
        help="The full original templates to compare the impulse response test with.",
    )
    group.add_option(
        "--impulse-bankno",
        metavar="index",
        action="store",
        help="The template bank index to perform the impulse test on.",
    )
    group.add_option(
        "--trigger-output",
        metavar="filename",
        action="append",
        default=[],
        help="Set the name of the LIGO light-weight XML output file *.{xml,xml.gz} "
             "or an SQLite database *.sqlite (required).  Can be given multiple times.  "
             "Exactly as many output files must be specified as svd-bank files will be "
             "processed (see --svd-bank).",
    )
    parser.add_option_group(group)

    group = OptionGroup(
        parser, "Ranking Statistic Options", "Adjust ranking statistic behaviour"
    )
    group.add_option(
        "--ranking-stat-output",
        metavar="filename",
        action="append",
        default=[],
        help="Set the name of the file to which to write ranking statistic data "
             "collected from triggers (optional).  Can be given more than once.  If given, "
             "exactly as many must be provided as there are --svd-bank options and they will"
             " be writen to in order.",
    )
    parser.add_option_group(group)

    group = OptionGroup(parser, "Program Behaviour")
    group.add_option(
        "--torch-device",
        action="store",
        default="cpu",
        help="The device to run LLOID and Trigger generation on.",
    )
    group.add_option(
        "--torch-dtype",
        action="store",
        type="str",
        default="float32",
        help="The data type to run LLOID and Trigger generation with.",
    )
    group.add_option(
        "-v", "--verbose", action="store_true", help="Be verbose (optional)."
    )
    group.add_option(
        "--output-kafka-server",
        metavar="addr",
        help="Set the server address and port number for output data. Optional",
    )
    group.add_option(
        "--analysis-tag",
        metavar="tag",
        default="test",
        help="Set the string to identify the analysis in which this job is part of. "
             'Used when --output-kafka-server is set. May not contain "." nor "-". Default '
             "is test.",
    )
    group.add_option(
        "--graph-name", metavar="filename", help="Plot pipieline graph to graph_name."
    )
    group.add_option(
        "--fake-sink", action="store_true", help="Connect to a FakeSeriesSink"
    )
    parser.add_option_group(group)

    options, args = parser.parse_args()

    return options, args


def main():
    # parse arguments
    options, args = parse_command_line()

    # sanity check options
    known_datasources = ["white", "sin", "impulse", "frames", "devshm"]
    if options.data_source in ["white", "sin", "impulse"]:
        if not options.num_buffers:
            raise ValueError(
                "Must specify num_buffers when data_source is one of 'white',"
                "'sin', 'impulse'"
            )
        elif not options.sample_rate:
            raise ValueError(
                "Must specify sample_rate when data_source is one of 'white',"
                "'sin', 'impulse'"
            )
        elif not options.source_buffer_duration:
            raise ValueError(
                "Must specify source_buffer_duration when data_source is one "
                "of 'white', 'sin', 'impulse'"
            )

        if options.data_source == "impulse":
            if not options.impulse_bank:
                raise ValueError("Must specify impulse_bank when data_source='impulse'")
            elif not options.impulse_bankno:
                raise ValueError(
                    "Must specify impulse_bankno when data_source='impulse'"
                )
            elif not options.impulse_ifo:
                raise ValueError("Must specify impulse_ifo when data_source='impulse'")

        source_ifos = None
    elif options.data_source == "frames":
        if not options.frame_cache:
            raise ValueError("Must specify frame_cache when data_source='frames'")
        elif not options.gps_start_time or not options.gps_end_time:
            raise ValueError(
                "Must specify gps_start_time and gps_end_time when "
                "data_source='frames'"
            )
        elif not options.channel_name:
            raise ValueError("Must specify channel_name when data_source='frames'")
        elif not options.sample_rate:
            # FIXME: shoud we just determine the sample rate from the gwf file?
            raise ValueError("Must specify sample_rate when data_source='frames'")
        elif not options.source_buffer_duration:
            raise ValueError("Must specify source_buffer_duration")
        channel_dict = parse_list_to_dict(options.channel_name)
        source_ifos = list(channel_dict.keys())
    elif options.data_source == "devshm":
        if not options.shared_memory_dir:
            raise ValueError("Must specify shared_memory_dir when data_source='devshm'")
        elif not options.channel_name:
            raise ValueError("Must specify channel_name when data_source='devshm'")
        options.source_buffer_duration = 1
        options.sample_rate = 16384
        channel_dict = parse_list_to_dict(options.channel_name)
        source_ifos = list(channel_dict.keys())
    else:
        raise ValueError(f"Unknown data source, must be one of {known_datasources}")

    # FIXME: currently track psd is always enabled in whitener
    if not options.reference_psd:
        options.track_psd = True

    # check pytorch data type
    dtype = options.torch_dtype
    if dtype == "float64":
        dtype = torch.float64
    elif dtype == "float32":
        dtype = torch.float32
    elif dtype == "float16":
        dtype = torch.float16
    else:
        raise ValueError("Unknown data type")

    # read in the svd banks
    banks = group_and_read_banks(
        svd_bank=options.svd_bank,
        source_ifos=source_ifos,
        nbank_pretend=options.nbank_pretend,
        nslice=options.nslice,
        verbose=True,
    )

    # sort and group the svd banks by sample rate
    sorted_bank = SortedBank(
        banks=banks,
        device=options.torch_device,
        dtype=dtype,
        nbank_pretend=options.nbank_pretend,
        nslice=options.nslice,
        verbose=True,
    )
    bank_metadata = sorted_bank.bank_metadata

    ifos = source_ifos
    maxrate = bank_metadata["maxrate"]

    #
    # Build pipeline
    #
    pipeline = Pipeline()

    # Data source
    source_out_links = datasource(pipeline, options, ifos)
    if options.data_source != "impulse":
        source_out_links, horizon_out_links, whiten_latency_out_links = condition(
            pipeline, options, ifos, maxrate, source_out_links
        )
    else:
        horizon_out_links = None

    num_samples = int(options.source_buffer_duration * maxrate)

    # connect LLOID
    lloid_output_source_link = lloid(
        pipeline,
        sorted_bank,
        source_out_links,
        num_samples,
        options.nslice,
        options.torch_device,
        dtype,
    )

    # make the sink
    if options.data_source == "impulse":
        pipeline.insert(
            ImpulseSink(
                name="imsink0",
                sink_pad_names=tuple(ifos) + (options.impulse_ifo + "_src",),
                original_templates=original_templates,
                template_duration=141,
                plotname="plots/response",
                impulse_pad=options.impulse_ifo + "_src",
                data_pad=options.impulse_ifo,
                bankno=options.impulse_bankno,
                verbose=options.verbose,
            ),
        )
        # link output of lloid
        for ifo, link in lloid_output_source_link.items():
            pipeline.insert(
                link_map={
                    "imsink0:sink:" + ifo: link,
                }
            )
            if ifo == impulse_ifo:
                pipeline.insert(
                    link_map={"imsink0:sink:" + ifo + "_src": source_out_links[ifo]}
                )
    else:
        # connect itacacac
        pipeline.insert(
            Itacacac(
                name="itacacac",
                sink_pad_names=tuple(ifos),
                source_pad_names=("trigs",),
                trigger_finding_length=options.trigger_finding_length,
                autocorrelation_banks=sorted_bank.autocorrelation_banks,
                template_ids=sorted_bank.template_ids,
                bankids_map=sorted_bank.bankids_map,
                end_times=sorted_bank.end_times,
                kafka=options.data_source == "devshm",
                device=options.torch_device,
            ),
        )
        for ifo in ifos:
            pipeline.insert(
                link_map={
                    "itacacac:sink:" + ifo: lloid_output_source_link[ifo],
                }
            )
        if options.data_source == "devshm":
            pipeline.insert(
                Latency(
                    name="itacacac_latency",
                    sink_pad_names=("data",),
                    source_pad_names=("latency",),
                    route="all_itacacac_latency",
                ),
                link_map={"itacacac_latency:sink:data": "itacacac:src:trigs"},
            )

        # Connect sink
        if options.fake_sink:
            pipeline.insert(
                FakeSeriesSink(
                    name="Sink",
                    sink_pad_names=("sink",),
                    verbose=options.verbose,
                ),
                link_map={"Sink:sink:sink": "itacacac:src:trigs"},
            )
        elif options.data_source == "devshm":
            pipeline.insert(
                KafkaSink(
                    name="KafkaSnk",
                    sink_pad_names=(
                                       "kafka",
                                       "itacacac_latency",
                                   )
                                   + tuple(ifo + "_whiten_latency" for ifo in ifos),
                    output_kafka_server=options.output_kafka_server,
                    topics=[
                               "gstlal." + options.analysis_tag + "." + ifo + "_snr_history"
                               for ifo in ifos
                           ]
                           + [
                               "gstlal." + options.analysis_tag + ".latency_history",
                           ]
                           + [
                               "gstlal." + options.analysis_tag + ".all_itacacac_latency",
                           ]
                           + [
                               "gstlal."
                               + options.analysis_tag
                               + "."
                               + ifo
                               + "_whitening_latency"
                               for ifo in ifos
                           ],
                    # topic="gstlal.greg_test.H1_snr_history",
                    routes=[ifo + "_snr_history" for ifo in ifos],
                    reduce_time=1,
                    verbose=options.verbose,
                ),
                link_map={
                    "KafkaSnk:sink:kafka": "itacacac:src:trigs",
                    "KafkaSnk:sink:itacacac_latency": "itacacac_latency:src:latency",
                },
            )
            for ifo in ifos:
                pipeline.insert(
                    link_map={
                        "KafkaSnk:sink:"
                        + ifo
                        + "_whiten_latency": whiten_latency_out_links[ifo],
                    }
                )
        else:
            pipeline.insert(
                StrikeSink(
                    name="StrikeSnk",
                    sink_pad_names=("trigs",)
                                   + tuple(["horizon_" + ifo for ifo in ifos]),
                    ifos=ifos,
                    verbose=options.verbose,
                    all_template_ids=sorted_bank.template_ids.numpy(),
                    bankids_map=sorted_bank.bankids_map,
                    subbankids=sorted_bank.subbankids,
                    template_sngls=sorted_bank.sngls,
                    trigger_output=options.trigger_output,
                    ranking_stat_output=options.ranking_stat_output,
                ),
                link_map={
                    "StrikeSnk:sink:trigs": "itacacac:src:trigs",
                },
            )
            for ifo in ifos:
                pipeline.insert(
                    link_map={
                        "StrikeSnk:sink:horizon_" + ifo: horizon_out_links[ifo],
                    }
                )

    # Plot pipeline
    if options.graph_name:
        pipeline.visualize(options.graph_name)

    # Run pipeline
    pipeline.run()

    #
    # Cleanup template bank temp files
    #

    print("shutdown: template bank cleanup", flush=True, file=sys.stderr)
    for ifo in banks:
        for bank in banks[ifo]:
            if options.verbose:
                print("removing file: ", bank.template_bank_filename, file=sys.stderr)
            os.remove(bank.template_bank_filename)

    print("shutdown: del bank", flush=True, file=sys.stderr)
    del bank


if __name__ == "__main__":
    main()



"""Sphinx configuration
"""

import dataclasses
import inspect
import operator
import os
import sys
import traceback

sys.path.insert(0, os.path.abspath("../src/"))

project = "SGN"
author = "Chad Hanna, Patrick Godwin, Jameson Rollins, James Kennington"
copyright = "2024, IGWN"
_repo_root_url = "https://git.ligo.org/greg/sgn"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.linkcode",
    "myst_parser",
]

html_theme = "furo"
html_title = "SGN Docs"


def linkcode_resolve(domain, info) -> str:
    """Function to resolve links to source code for use with the linkcode extension.

    Args:
        domain:
        info:

    References:
        [1] LinkCode Extension: https://www.sphinx-doc.org/en/master/usage/extensions/linkcode.html
        [2] scikit-learn linkcode_resolve: https://github.com/scikit-learn/scikit-learn/blob/f7026e575a494b
                                           47b50557932b5c3ce0688fdc72/doc/sphinxext/github_link.py
        [3] Issue with DataClass attrs: https://stackoverflow.com/questions/77848299/operator-attrgetter-
                                        cannot-read-attributes-in-a-dataclass

    Returns:
        str: URL to source code
    """
    if domain not in ("py", "pyx"):
        return
    if not info.get("module") or not info.get("fullname"):
        return

    try:
        obj_tokens = info["fullname"].split(".")
        class_name = obj_tokens[0]
        module = __import__(info["module"], fromlist=[class_name])

        # Check for dataclass, non-defaulted fields will not be class attrs and will not be importable
        # directly from the type, so we need to get the class object and check the fields
        class_obj = getattr(module, class_name)
        if dataclasses.is_dataclass(class_obj) and len(obj_tokens) > 1:
            fields = dataclasses.fields(class_obj)
            field_dict = {field.name: field for field in fields}
            field = field_dict.get(info["fullname"].split(".")[1])

            # If the field has no default, we can only locate the class, otherwise we can locate the field
            if field is None or field.default is dataclasses.MISSING:
                obj = class_obj
            else:
                obj = operator.attrgetter(info["fullname"])(module)
        else:
            obj = operator.attrgetter(info["fullname"])(module)

        # Unwrap the object to get the correct source
        # file in case that is wrapped by a decorator
        obj = inspect.unwrap(obj)
    except Exception as e:
        # Print formatted traceback, used for debugging
        formatted_traceback = "".join(
            traceback.format_exception(type(e), e, e.__traceback__)
        )
        print(formatted_traceback)
        return

    # Extract the filename
    try:
        fn = inspect.getsourcefile(obj)
    except Exception:
        fn = None
    if not fn:
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except Exception:
            fn = None
    if not fn:
        return

    fn = os.path.relpath(fn, start=os.path.dirname(__import__("sgn").__file__))

    # Try to get line number
    try:
        lineno = inspect.getsourcelines(obj)[1]
    except Exception:
        lineno = ""

    if lineno:
        lineno = f"#L{lineno}"

    return f"{_repo_root_url}/-/blob/main/src/sgn/{fn}{lineno}"



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any

import copy
import numpy
import os
import sys
import warnings
import scipy
import tempfile

import lal

from ligo.lw import ligolw
from ligo.lw import lsctables
from ligo.lw import array as ligolw_array
from ligo.lw import param as ligolw_param
from ligo.lw import utils as ligolw_utils
from ligo.lw.utils import process as ligolw_process

from .psd import condition_psd, HorizonDistance

Attributes = ligolw.sax.xmlreader.AttributesImpl


class DefaultContentHandler(ligolw.LIGOLWContentHandler):
    pass


ligolw_array.use_in(DefaultContentHandler)
ligolw_param.use_in(DefaultContentHandler)
lsctables.use_in(DefaultContentHandler)


@dataclass
class BankFragment:
    rate: int
    start: float
    end: float
    # orthogonal_template_bank: Sequence[Any]
    # singular_values: Sequence[Any]
    # mix_matrix: Sequence[Any]
    # chifacs: Sequence[Any]
    # sum_of_squares_weights: Sequence[Any]


@dataclass
class Bank:
    sngl_inspiral_table: Sequence[Any]
    filter_length: float
    gate_threshold: float
    logname: str
    snr_threshold: float
    template_bank_filename: str
    bank_id: str
    bank_type: str
    autocorrelation_bank: Sequence[Any]
    autocorrelation_mask: Sequence[Any]
    sigmasq: Sequence[Any]
    bank_correlation_matrix: Sequence[Any]
    horizon_factors: dict[int, float]
    processed_psd: Sequence[Any]
    bank_fragments: list[Any]

    def get_rates(self):
        return set(bank_fragment.rate for bank_fragment in self.bank_fragments)


def preferred_horizon_distance_template(banks):
    template_id, m1, m2, s1z, s2z = min(
        (row.template_id, row.mass1, row.mass2, row.spin1z, row.spin2z)
        for bank in banks
        for row in bank.sngl_inspiral_table
    )
    return template_id, m1, m2, s1z, s2z


def horizon_distance_func(banks):
    """
    Takes a dictionary of objects returned by read_banks keyed by instrument
    """
    # span is [15 Hz, 0.85 * Nyquist frequency]
    # find the Nyquist frequency for the PSD to be used for each
    # instrument.  require them to all match
    nyquists = set((max(bank.get_rates()) / 2.0 for bank in banks))
    if len(nyquists) != 1:
        warnings.warn(
            "all banks should have the same Nyquist frequency to define a consistent horizon distance function (got %s)"
            % ", ".join("%g" % rate for rate in sorted(nyquists))
        )
    # assume default 4 s PSD.  this is not required to be correct, but
    # for best accuracy it should not be larger than the true value and
    # for best performance it should not be smaller than the true
    # value.
    deltaF = 1.0 / 4.0
    # use the minimum template id as the cannonical horizon function
    template_id, m1, m2, s1z, s2z = preferred_horizon_distance_template(banks)

    return template_id, HorizonDistance(
        15.0,
        0.85 * max(nyquists),
        deltaF,
        m1,
        m2,
        spin1=(0.0, 0.0, s1z),
        spin2=(0.0, 0.0, s2z),
    )


def read_banks(filename, contenthandler, verbose=False):
    """Read SVD banks from a LIGO_LW xml file."""

    # Load document
    xmldoc = ligolw_utils.load_url(
        filename, contenthandler=contenthandler, verbose=verbose
    )

    banks = []

    # FIXME in principle this could be different for each bank included in
    # this file, but we only put one in the file for now
    # FIXME, right now there is only one instrument so we just pull out the
    # only psd there is
    try:
        raw_psd = list(lal.series.read_psd_xmldoc(xmldoc).values())[0]
    except ValueError:
        # the bank file does not contain psd ligolw element.
        raw_psd = None

    for root in (
        elem
        for elem in xmldoc.getElementsByTagName(ligolw.LIGO_LW.tagName)
        if elem.hasAttribute("Name") and elem.Name == "gstlal_svd_bank_Bank"
    ):

        # Create new SVD bank object
        bank = Bank.__new__(Bank)

        # Read sngl inspiral table
        bank.sngl_inspiral_table = lsctables.SnglInspiralTable.get_table(root)
        bank.sngl_inspiral_table.parentNode.removeChild(bank.sngl_inspiral_table)

        # Read root-level scalar parameters
        bank.filter_length = ligolw_param.get_pyvalue(root, "filter_length")
        bank.gate_threshold = ligolw_param.get_pyvalue(root, "gate_threshold")
        bank.logname = ligolw_param.get_pyvalue(root, "logname") or None
        bank.snr_threshold = ligolw_param.get_pyvalue(root, "snr_threshold")
        bank.template_bank_filename = ligolw_param.get_pyvalue(
            root, "template_bank_filename"
        )
        bank.bank_id = ligolw_param.get_pyvalue(root, "bank_id")
        bank.bank_type = ligolw_param.get_pyvalue(root, "bank_type")

        try:
            bank.newdeltaF = ligolw_param.get_pyvalue(root, "new_deltaf")
            bank.working_f_low = ligolw_param.get_pyvalue(root, "working_f_low")
            bank.f_low = ligolw_param.get_pyvalue(root, "f_low")
            bank.sample_rate_max = ligolw_param.get_pyvalue(root, "sample_rate_max")
        except ValueError:
            pass

        # Read root-level arrays
        bank.autocorrelation_bank = (
            ligolw_array.get_array(root, "autocorrelation_bank_real").array
            + 1j * ligolw_array.get_array(root, "autocorrelation_bank_imag").array
        )
        bank.autocorrelation_mask = ligolw_array.get_array(
            root, "autocorrelation_mask"
        ).array
        bank.sigmasq = ligolw_array.get_array(root, "sigmasq").array
        bank_correlation_real = ligolw_array.get_array(
            root, "bank_correlation_matrix_real"
        ).array
        bank_correlation_imag = ligolw_array.get_array(
            root, "bank_correlation_matrix_imag"
        ).array
        bank.bank_correlation_matrix = (
            bank_correlation_real + 1j * bank_correlation_imag
        )

        # prepare the horizon distance factors
        bank.horizon_factors = dict(
            (row.template_id, sigmasq**0.5)
            for row, sigmasq in zip(bank.sngl_inspiral_table, bank.sigmasq)
        )

        if raw_psd is not None:
            # reproduce the whitening psd and attach a reference to the psd
            bank.processed_psd = condition_psd(
                raw_psd,
                bank.newdeltaF,
                minfs=(bank.working_f_low, bank.f_low),
                maxfs=(bank.sample_rate_max / 2.0 * 0.90, bank.sample_rate_max / 2.0),
            )
        else:
            bank.processed_psd = None

        # Read bank fragments
        bank.bank_fragments = []
        for el in (
            node for node in root.childNodes if node.tagName == ligolw.LIGO_LW.tagName
        ):
            frag = BankFragment(
                rate=ligolw_param.get_pyvalue(el, "rate"),
                start=ligolw_param.get_pyvalue(el, "start"),
                end=ligolw_param.get_pyvalue(el, "end"),
            )

            # Read arrays
            frag.chifacs = ligolw_array.get_array(el, "chifacs").array
            try:
                frag.mix_matrix = ligolw_array.get_array(el, "mix_matrix").array
            except ValueError:
                frag.mix_matrix = None
            frag.orthogonal_template_bank = ligolw_array.get_array(
                el, "orthogonal_template_bank"
            ).array
            try:
                frag.singular_values = ligolw_array.get_array(
                    el, "singular_values"
                ).array
            except ValueError:
                frag.singular_values = None
            try:
                frag.sum_of_squares_weights = ligolw_array.get_array(
                    el, "sum_of_squares_weights"
                ).array
            except ValueError:
                frag.sum_of_squares_weights = None
            bank.bank_fragments.append(frag)

        banks.append(bank)
    template_id, func = horizon_distance_func(banks)
    template_id = abs(
        template_id
    )  # make sure horizon_distance_func did not pick the noise model template
    horizon_norm = None
    for bank in banks:
        if template_id in bank.horizon_factors and bank.bank_type == "signal_model":
            assert horizon_norm is None
            horizon_norm = bank.horizon_factors[template_id]
    for bank in banks:
        bank.horizon_distance_func = func
        bank.horizon_factors = dict(
            (tid, f / horizon_norm) for (tid, f) in bank.horizon_factors.items()
        )
    xmldoc.unlink()
    return banks


def parse_bank_files(svd_banks, verbose, snr_threshold=None):
    """
    given a dictionary of lists of svd template bank file names parse them
    into a dictionary of bank classes
    """

    banks = {}

    for instrument, filename in svd_banks.items():
        for n, bank in enumerate(
            read_banks(filename, contenthandler=DefaultContentHandler, verbose=verbose)
        ):
            # Write out sngl inspiral table to temp file for
            # trigger generator
            # FIXME teach the trigger generator to get this
            # information a better way
            bank.template_bank_filename = tempfile.NamedTemporaryFile(
                suffix=".xml.gz", delete=False
            ).name
            xmldoc = ligolw.Document()
            # FIXME if this table reference is from a DB this
            # is a problem (but it almost certainly isn't)
            xmldoc.appendChild(ligolw.LIGO_LW()).appendChild(
                bank.sngl_inspiral_table.copy()
            ).extend(bank.sngl_inspiral_table)
            ligolw_utils.write_filename(
                xmldoc, bank.template_bank_filename, verbose=verbose
            )
            xmldoc.unlink()  # help garbage collector
            bank.logname = "%sbank%d" % (instrument, n)
            banks.setdefault(instrument, []).append(bank)
            if snr_threshold is not None:
                bank.snr_threshold = snr_threshold

    # FIXME remove when this is no longer an issue
    if not banks:
        raise ValueError(
            "Could not parse bank files into valid bank dictionary.\n\t- Perhaps you are using out-of-date svd bank files?  Please ensure that they were generated with the same code version as the parsing code"
        )
    return banks
